/* Copyright (c) 2003-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Komodo's Debug Session tab handling code.
 *
 * Here is the deal: When you start a debug session in Komodo you get a new
 * Debug Session Tab in the bottom pane. This is it here. The Komodo code that
 * starts a debug session calls into here to get a new tab. A handler XPCOM
 * component (DebugSessionTabManager) is created for the tab and Komodo's
 * debugging works with the debug tab through its API.
 *
 * Usage:
 *      this.manager.getTabMgr(
 *          function onSuccessCallback(tabMgr) {
 *              tabMgr.configure(session);
 *              tabMgr.sessionInit();
 *              tabMgr.sessionStart();
 *              tabMgr.sessionEnd();
 *              ...use the koIDebugSessionTabManager interface...
 *          },
 *          function onErrorCallback() {
 *              // Failed to create a tab and manager. The user was notified via
 *              // an alert. The debug session should be aborted.
 *          }
 *      );
 *
 */

//---- globals


if (typeof(ko) == 'undefined') {
    var ko = {};
}
if (typeof(ko.dbg) == 'undefined') {
    ko.dbg = {};
}

(function() { /* ko.dbg */

const MAX_DEBUG_SESSION_TABS = 100;
var log = ko.logging.getLogger("debugSessionTab");
//log.setLevel(ko.logging.LOG_DEBUG);
const NS_XUL = 'http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul';
const Cc = Components.classes;
const Ci = Components.interfaces;
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

// Cache of tab managers.
// - Maps a tab number/id to a DebugSessionTabManager() instance.
// - If managers[<tab-number>] does not exist then no such tab/tabpanel has
//   been created.
// - If managers[<tab-number>]==null then there is a tab/tabpanel left over
//   from a previous debug session, it is just collapsed.
// - If managers[<tab-number>]==<DebugSessionTabManager-instance> then
//   there is an existing debug session in that tab/tabpanel.
//   XXX Note that this debug session may no longer be active in which
//       case we *may* want to re-use this tab. Hmmm. How about adding a
//       "pin" icon to an inactive tab to ensure it doesn't get re-used.
//       Could do the same for the Find Results windows.


//---- the DebugSessionTab component

function DebugSessionTabManager(id) {
    this._outputTmpFile = null;

    // Number identifying this DebugSessionTabManager from others.
    this.id = id;
    this._idprefix = "debugsession"+this.id;

    this.name = null;
    this.contextNames = null;
    // Cache of useful widget references in this tab.
    this.tab = document.getElementById(this._idprefix+"-tab");
    this.tabPanel = document.getElementById(this._idprefix+"-tabpanel");

    // the rest needs the panel to finish loading before we can use it; do
    // that in .initialize()
}
DebugSessionTabManager.prototype.constructor = DebugSessionTabManager;

DebugSessionTabManager.prototype.QueryInterface =
    XPCOMUtils.generateQI([Ci.koIDebugSessionTabManager]);

DebugSessionTabManager.prototype.initialize = function DebugSessionTabManager_initialize(id)
{
    log.debug("DebugSessionTabManager["+id+"].initialize()");
    try {
        this.tabPanel.contentDocument
            .addEventListener("ko-widget-move-completed", (function(event) {
                // the widget moved, update the references
                this.tab = document.getElementById(this._idprefix+"-tab");
                this.tabPanel = document.getElementById(this._idprefix+"-tabpanel");
            }).bind(this), false);
        this.debuggerPanel = this.tabPanel.contentDocument.getElementById("debuggerPanel");
        this.debuggerPanel.onLoad(this, true);

        this._watched = {};
        this._varState = {};
        this.currentStackFrameIndex = -1;
        this.resetTabName();
        this.clear();
        this.isInteractive = false;
        this.basedir = null;

        /**
         * {boolean} Acts as a latch for when the session is about to go from
         *      running to stopped (possibly by evaluating interactive mode).
         *      This is used for variable change highlighting.
         */
        this._sessionEnteringBreak = false;
    } catch(ex) {
        log.exception(ex);
        Components.utils.reportError(ex);
    }
}

DebugSessionTabManager.prototype.initVarTabs = function DebugSessionTabManager_initVarTabs(/* XXX arguments here? */)
{
    log.debug("DebugSessionTabManager["+this.id+"].initVarTabs(...)");
    try {
        this.debuggerPanel.watchVariablesTree.onLoad(new variableTreeView(this, 'Watched'));

        // get the context names now
        if (!this.contextNames) {
            this.contextNames = this.session.dbg.contextNames();
        }

        // XXX for debugging only
        log.debug("  number of contexts: "+this.contextNames.length);
        if (this.contextNames.length < 0) return;

        // set up the trees
        // we do not include the watched vars tab here, as this stuff
        // needs to be dynamic
        this._varTabs = {};
        for (var i = 0; i < this.contextNames.length; i++) {
            log.debug("  context: "+this.contextNames[i].name);
            var contextNameField = this.contextNames[i];
            var contextNameValue = contextNameField.name;
            var contextTab = this.debuggerPanel.addVariablePanel(contextNameField);
            var treeView = new variableTreeView(this, contextNameValue);
            contextTab[1].onLoad(treeView);
            this._varTabs[contextNameValue] = contextTab;
            treeView.setupInitialSortDirection(this.session.dbg.languageName,
                                               contextTab[1]);
        }

    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.configure = function(session)
{
    log.debug("DebugSessionTabManager["+this.id+"].configure(...)");
    try {
        if (typeof(session) == 'undefined')
            session = null;
        this.session = session;
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.show = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].show()");
    try {
        // If necessary, open output pane in komodo.xul
        ko.uilayout.ensureOutputPaneShown();

        // switch to proper tab
        var pane = document.getElementById(this._idprefix+"-tabpanel");
        pane.tabbox.selectedTab = pane.tab;
    } catch(ex) {
        log.exception(ex);
    }
}


DebugSessionTabManager.prototype.clear = function DebugSessionTabManager_clear()
{
    log.debug("DebugSessionTabManager["+this.id+"].clear()");
    try {
        this.debuggerPanel.terminalView.clear();
    } catch(ex) {
        log.exception(ex);
    }
}


DebugSessionTabManager.prototype.canClose = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].canClose()");
    try {
        // Warn if this tab is still active, offer to stop it if so.
        if (!this.session) return true;

        var sessionType = 'debugging session';
        var doNotAskPref = 'debugger_confirm_close';
        if (this.session.isInteractive) {
            sessionType = 'interactive shell';
            doNotAskPref = 'interactiveshell_confirm_close';
        }

        var answer = ko.dialogs.customButtons(
            "This "+sessionType+" is still in progress. The session "+
            "must be stopped before the tab can be closed.",
            ["&Kill Session and Close Tab", "Cancel"],
            "Cancel",  // default button
            null, // text
            null, // title
            doNotAskPref);
        if (answer == "Kill Session and Close Tab") {
            return true;
        } else { // answer == "Cancel"
            return false;
        }
    } catch(ex) {
        log.exception(ex);
    }
    return false;
}

DebugSessionTabManager.prototype.forceClose = function DebugSessionTabManager_forceClose()
{
    log.debug("DebugSessionTabManager["+this.id+"].forceClose()");
    try {
        if (this.session) {
            this.session.terminate();
        }
        this.debuggerPanel.releaseTerminal();
        this.removeHTMLTempFile();
    } catch(e) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.resetTabName = function()
{
    var name = 'Output';
    if (this.session) {
        if (this.session.invoke_type == ko.dbg.invocation.INVOKE_INTERACTIVE)
            name = this.session.dbg.languageName+' Shell';
        else if (this.session.invoke_type == ko.dbg.invocation.INVOKE_RUN)
            name = "Run Output";
        else if (this.session.invoke_type == ko.dbg.invocation.INVOKE_DEBUG)
            name = "Debug Output";
    }
    this.tabPanel.setAttribute('label', name);
}

/* sessionInit starts prior to having a debugger connection,
   sessionStart occurs after a having a debugger connection.
   This is important becuase prior to connection we need to have
   some things in place, such as the terminalHandler calling
   startSession to configure io.  At this point the session.dbg
   variable is NOT available.  Since some items may need this for
   initialization (such as the variable viewers), we now have
   these two functions.  */
DebugSessionTabManager.prototype.sessionInit = function DebugSessionTabManager_sessionInit(isInteractive)
{
    log.debug("DebugSessionTabManager["+this.id+"].sessionInit()");
    try {
        this.debuggerPanel.terminalView.clear();
        this.debuggerPanel.terminalView.startSession();
        this.isInteractive = isInteractive;
        this.debuggerPanel.setInteractive(isInteractive);
        this.name = null;
        this.resetTabName();
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.sessionStart = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].startSession()");
    try {
        // startup items that require access to session.dbg
        this.initVarTabs();
        this.resetTabName();
        this.loadHTMLView();
        this.basedir = null;
        if (this.session && this.session.filename) {
            var osPathSvc = Components.classes["@activestate.com/koOsPath;1"].
                  getService(Components.interfaces.koIOsPath);
            this.basedir = osPathSvc.dirname(this.session.filename);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.sessionEnd = function DebugSessionTabManager_sessionEnd()
{
    log.debug("DebugSessionTabManager["+this.id+"].sessionEnd()");
    try {
        this.debuggerPanel.terminalView.endSession();
        this.clearThreadMenu();
        this.clearCallStackList();
        this.clearVariables();
        this.debuggerPanel.removeVariablePanels();
        this.session = null;
        ko.dbg.tabManager.releaseTabMgr(this);
        this.currentStackFrameIndex = -1;
        this.updateWatchedVars();
        if (this.isInteractive) {
            if (ko.prefs.getBooleanPref("interactiveShellCloseOnEnd")) {
                // Close interactive session tabs when done
                ko.dbg.tabManager.closeTab(this.tab);
                // focus in the editor.
                if (ko.views.manager.currentView) {
                    ko.views.manager.currentView.setFocus();
                }
            }
        }
        this.isInteractive = false;
        this.contextNames = null;
    } catch(ex) {
        log.exception(ex);
    }
}

/**
 * Called when the state of the session changes (e.g. run->break, or entering
 * interactive mode).
 * 
 * @param   {Number} state The new session state, as a koIDBGPSession::STATUS_*
 *              constant.
 */
DebugSessionTabManager.prototype.sessionStateChange = function DebugSessionTabManager_sessionStateChange(state)
{
    switch (state) {
        case ko.dbg.manager.STATE_BREAK:
        case ko.dbg.manager.STATE_INTERACTIVE:
            this._sessionEnteringBreak = true;
            break;
        default:
            this._sessionEnteringBreak = false;
    }
};

DebugSessionTabManager.prototype.clearVariables = function() {
    log.debug("DebugSessionTabManager["+this.id+"].clearVariables()");
    try {
        for (var context in this._varTabs) {
            var varTreeView = this._varTabs[context][1].view;
            varTreeView.savePreferences();
            varTreeView.clearList();
        }
    } catch(ex) {
        log.exception(ex);
    }
    delete this._varTabs;
}

/**
 * Update the variables view associated with this debug session
 * 
 * @see variableTreeView.updateList()
 */
DebugSessionTabManager.prototype.updateVariables = function DSTM_updateVariables() {
    log.debug("DebugSessionTabManager["+this.id+"].updateVariables()");
    try {
        if (!this.session) return;
        var state = this.session.getState();
        if (state != ko.dbg.manager.STATE_BREAK && state != ko.dbg.manager.STATE_INTERACTIVE) {
            return;
        }
        // update information displayed about variables defined in the
        // current stack frame

        // only update variables in the currently visible tab
        var tab = this.debuggerPanel.variablesTabs.selectedItem;
        var currentContext = tab.getAttribute('label');

        if (!(currentContext in this._varTabs)) {
            this.updateWatchedVars();
            return;
        }
        var varTab = this._varTabs[currentContext][1];
        // get the current selection so we can reset that
        // after updating the variable list
        var ci = varTab.view.mSelection.currentIndex;
        
        // get the current context names
        if (!this.contextNames) {
            this.contextNames = this.session.dbg.contextNames();
        }

        // find the visible context to update
        var context = this.contextNames.filter(function(cx) cx.name == currentContext).pop();

        if (context) {
            var tmp = new Object();
            this.session.dbg.contextGet(context.id,
                                        this.currentStackFrameIndex,
                                        tmp, new Object());
            var debugProps = tmp.value;
            varTab.view.updateList(debugProps, this._sessionEnteringBreak);
            this._sessionEnteringBreak = false;
        } else {
            varTab.view.clearList();
        }

        varTab.view.selectIndex(ci);
    } catch(ex) {
        log.exception(ex);
    }
    return;
}

DebugSessionTabManager.prototype.updateWatchedVars = function DebugSessionTabManager_updateWatchedVars() {
    log.debug("DebugSessionTabManager["+this.id+"].updateWatchedVars()");
    try {
        if (!this.debuggerPanel.watchVariablesTree.view) return;
        // clear the expression variables
        var vars = [];
        var expression;
        // get them again
        if (this.session) {
            var curr_dbg_session = this.session.dbg;
            var languageName = curr_dbg_session.languageName;
            for (expression in this._watched) {
                log.debug('+++ showing expression ['+expression+']\n');
                var p = curr_dbg_session.propertyGetEx(0, this.currentStackFrameIndex,
                                                       expression, 0, 0, 0, "");
                if (ko.dbg.is_exception_property(p, languageName)) {
                    var p_alt = curr_dbg_session.evalString(expression);
                    if (!ko.dbg.is_exception_property(p_alt, languageName)) {
                        p = p_alt;
                    }
                }
                this._watched[expression] = p;
                vars.push(p);
            }
        } else {
            for (expression in this._watched) {
                log.debug('--- showing expression ['+expression+']\n');
                this._watched[expression] = null;
                vars.push(expression);
            }
        }
        // display them
        this.debuggerPanel.watchVariablesTree.view.updateList(vars, this._sessionEnteringBreak);
        this._sessionEnteringBreak = false;
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.addWatchedVar = function(expression, value) {
    log.debug("DebugSessionTabManager["+this.id+"].addWatchedVar("+expression+")");
    try {
        this._watched[expression] = null;
        if (this.session) {
            if (typeof(value) != 'undefined') {
                this.session.dbg.propertySet(expression, value);
            }
        }
        this.updateWatchedVars();
    } catch(ex) {
        log.exception(ex);
    }
}


DebugSessionTabManager.prototype.removeWatchedVarName = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].removeWatchedVarName("+expression+")");
    try {
        if (typeof(this._watched[expression]) != 'undefined') {
            delete this._watched[expression];
            return true;
        }
    } catch(ex) {
        log.exception(ex);
    }
    return false;
}

DebugSessionTabManager.prototype.haveWatchedVar = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].haveWatchedVar("+expression+")");
    try {
        var haveit = (typeof(this._watched[expression]) != 'undefined')
    } catch(ex) {
        log.exception(ex);
    }
    return haveit;
}

/**
 * take the text from the debugger output scintilla, put it into the
 * iframe, and switch
 */
DebugSessionTabManager.prototype.removeHTMLTempFile = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].removeHTMLTempFile()");
    try {
        if (this._outputTmpName) {
            var tmpFileSvc = Components.classes["@activestate.com/koFileService;1"]
                             .getService(Components.interfaces.koIFileService)
            tmpFileSvc.deleteTempFile(this._outputTmpName, true);
            this._outputTmpName = null;
        }
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.loadHTMLView = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].renderOutputAsHTML()");
    try {
        this.removeHTMLTempFile();
        var doc;
        var osPathSvc = Components.classes["@activestate.com/koOsPath;1"].
                  getService(Components.interfaces.koIOsPath);
        var docSvc = Components.classes['@activestate.com/koDocumentService;1']
                    .getService(Components.interfaces.koIDocumentService);

        // get the text from scintilla, stuff it into the iframe, and switch the deck
        var text = this.debuggerPanel.terminalView.scimoz.text;
        if (text) {
            var tmpFileSvc = Components.classes["@activestate.com/koFileService;1"]
                             .getService(Components.interfaces.koIFileService)

            if (this.basedir) {
                this._outputTmpName = tmpFileSvc.makeTempNameInDir(this.basedir,".html");
            } else {
                this._outputTmpName = tmpFileSvc.makeTempName(".html");
            }

            doc = docSvc.createDocumentFromURI(this._outputTmpName);
            doc.setBufferAndEncoding(text, this.debuggerPanel.terminalView.scintilla.encoding);
            doc.save(true);
            this.debuggerPanel.browser.initWithDocument(doc, false);
        } else {
            this.debuggerPanel.browser.open('about:blank', false);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.updateHTMLView = function dbgSessionTabMgr_updateHTMLView()
{
    log.debug("DebugSessionTabManager["+this.id+"].updateHTMLView()");
    try {
        if (this._outputTmpName) {
            this.removeHTMLTempFile();
        }
        this.loadHTMLView();
    } catch(ex) {
        log.exception(ex);
    }
}

/*
  called when the state of the session changes
*/
DebugSessionTabManager.prototype.updateUI = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].updateUI()");
    try {
        var state = ko.dbg.manager.STATE_STOPPED;
        if (this.session) {
            state = this.session.getState();
            if (this.session.dbg)
                this.tabPanel.setAttribute('dbg_language', this.session.dbg.languageName);
        }
        log.debug("DebugSessionTabManager[" + this.id + "].updateUI(): state=" +
                  state + "(" + ko.dbg.tabManager.getStatusAsString(state) + ")" +
                  " session=" + this.session);
        this.tabPanel.setAttribute('dbg_status', ko.dbg.tabManager.getStatusAsString(state));

        if (this.session && ko.dbg.manager.currentSession === this.session) {
            this.tab.setAttribute("current-session", true);
        } else {
            this.tab.removeAttribute("current-session");
        }

        // always update our thread and call stack.  We'll only refill the callstack
        // if the current session is in break mode.  There is the potential that a thread
        // will change a value on us, but not much we can do about that without a lot
        // more overhead
        if (!this.isInteractive)
            this.buildThreadMenu();

        this.clearCallStackList();

        switch(state) {
        case ko.dbg.manager.STATE_STOPPED:
            this.debuggerPanel.infoText.value = "Debugging session has ended.";
            break;

        case ko.dbg.manager.STATE_STARTING:
            this.debuggerPanel.infoText.value = "Debugger is starting...";
            break;

        case ko.dbg.manager.STATE_RUNNING:
            this.debuggerPanel.infoText.value = "Debugger is running...";
            break;

        case ko.dbg.manager.STATE_INTERACTIVE:
            if (this.isInteractive) {
                if (!this.debuggerPanel.getElement('vbox','varViewers').getAttribute('collapsed'))
                    this.updateVariables();
                break;
            }
        case ko.dbg.manager.STATE_BREAK:
            this.buildCallStackList();

            var frame, haveFrame = false;
            if (this.stackFrames) {
                frame = this.stackFrames[0];
                if (frame)
                    haveFrame = true;
            }
            if (haveFrame)
                this.debuggerPanel.infoText.value =  "Break at " +
                                ko.uriparse.displayPath(frame.filename)
                                + ", line " + frame.lineno + ".";
            else
                this.debuggerPanel.infoText.value =  "";
            break;

        case ko.dbg.manager.STATE_STOPPING:
            this.debuggerPanel.infoText.value = "Debugger is stopping...";
            break;

        } // end switch(new_state)
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.clearThreadMenu = function DebugSessionTabManager_clearThreadMenu()
{
    log.debug("DebugSessionTabManager["+this.id+"].clearThreadMenu()");
    try {
        /* clear the thread list */
        this.threads = null;
        while (this.debuggerPanel.threadsMenuPopup.childNodes.length > 0) {
            this.debuggerPanel.threadsMenuPopup.removeChild(this.debuggerPanel.threadsMenuPopup.lastChild);
        }
        this.debuggerPanel.threadsBox.setAttribute('collapsed',true);
        this.debuggerPanel.threadsMenuList.value = null;
    } catch(ex) {
        log.exception(ex);
    }
}

/* reset the threads list */
DebugSessionTabManager.prototype.buildThreadMenu = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].buildThreadMenu()");
    try {
        this.clearThreadMenu();
        if (!this.session || !this.session.application)
            return;

        var tmp = new Object();
        this.session.application.getSessionList(tmp, new Object());
        this.threads = tmp.value;
        if (this.threads.length < 2)
            return;

        this.debuggerPanel.threadsBox.removeAttribute('collapsed');
        var item, thread;
        for (var i = 0; i < this.threads.length; ++i) {
            thread = this.threads[i];
            item = document.createElementNS(NS_XUL, 'menuitem');
            item.setAttribute('class','menuitem-iconic dbg_status_icon');
            item.setAttribute('label', thread.threadId);
            item.setAttribute('dbg_status', ko.dbg.tabManager.getStatusAsString(thread.status));
            item.setAttribute('dbg_language', thread.languageName);
            item.setAttribute('oncommand','this.value.selectSessionThread('+i+');');
            item.setAttribute('flex','1');
            item.value = this;
            this.debuggerPanel.threadsMenuPopup.appendChild(item);
            if (thread == this.session.dbg) {
                this.debuggerPanel.threadsMenuList.selectedIndex = i;
            }
        }
        if (this.debuggerPanel.threadsMenuList.selectedIndex < 0)
            this.debuggerPanel.threadsMenuList.selectedIndex = 0;
        this.currentThreadIndex = this.debuggerPanel.threadsMenuList.selectedIndex;
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.selectSessionThread = function(index)
{
    /* this is to allow the thread dropdown menu to release, since selecting
        a different thread can result in rebuilding the menu also */
    this._timeout = new ko.objectTimer(this,this._selectSessionThread,[index]);
    this._timeout.startTimeout(1);
}

DebugSessionTabManager.prototype._selectSessionThread = function(index)
{
    log.debug("DebugSessionTabManager["+this.id+"].selectSessionThread("+index+")");
    try {
        if (index == this.currentThreadIndex) {
            log.debug("  already selected: "+index);
            return;
        }
        this.currentThreadIndex = index;
        this.session.application.currentSession = this.threads[index];
        if (this.session.getState() == ko.dbg.manager.STATE_RUNNING) {
            // cannot do stack for running threads
            this.clearCallStackList();
            return;
        }
        // reset the UI for the new thread selection
        this.session.updateUI();
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.clearCallStackList = function DebugSessionTabManager_clearCallStackList()
{
    log.debug("DebugSessionTabManager["+this.id+"].clearCallStackList()");
    try {
        /* clear the thread list */
        this.stackFrames = null;

        while (this.debuggerPanel.callstackListbox.childNodes.length > 1) {
            this.debuggerPanel.callstackListbox.removeChild(this.debuggerPanel.callstackListbox.childNodes[1]);
        }
    } catch(ex) {
        log.exception(ex);
    }
}

/**
 * Build and populate the call stack UI.
 *
 * Note: clearCallStackList should be called prior to this method.
 */
DebugSessionTabManager.prototype.buildCallStackList = function()
{
    log.debug("DebugSessionTabManager["+this.id+"].buildCallStackList()");
    try {
        var state = this.session.getState();
        if (state != ko.dbg.manager.STATE_BREAK && state != ko.dbg.manager.STATE_INTERACTIVE) {
            return;
        }
        var tmp = new Object();
        this.session.dbg.stackFramesGet(tmp, new Object());
        this.stackFrames = tmp.value;
        if (this.stackFrames.length < 1) {
            if (state == ko.dbg.manager.STATE_INTERACTIVE) {
                this.tabPanel.setAttribute('label', this.session.dbg.languageName+' Shell');
            }
            return;
        }
        if (state == ko.dbg.manager.STATE_INTERACTIVE) {
            this.tabPanel.setAttribute('label', this.session.dbg.languageName+' Shell');
        } else {
            this.name = ko.uriparse.baseName(this.stackFrames[0].filename);
            this.tabPanel.setAttribute('label', 'Debug: '+this.name);
        }

        var item, frame;
        for (var i = 0; i < this.stackFrames.length; ++i) {
            frame = this.stackFrames[i];
            var frameName = ko.uriparse.baseName(frame.filename) +
                        ', line ' + frame.lineno +
                        ', in ' + frame.where;
            item = this.debuggerPanel.callstackListbox.appendItem(frameName, this)
            item.setAttribute('style', 'margin: 0px; border:none');
            item.setAttribute('class','listitem-iconic');
            item.value = this;
        }
        this.debuggerPanel.callstackListbox.selectedIndex = 0;
        this.currentStackFrameIndex = 0;
        this.doStackFrame();

        this.debuggerPanel.callstackListbox.selectedItem.setAttribute('currentview','true');
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.selectStackFrame = function(index)
{
    log.debug("DebugSessionTabManager["+this.id+"].selectStackFrame("+index+")");
    try {
        if (index == this.currentStackFrameIndex) {
            log.debug("  already selected: "+index);
            return;
        }
        if (this.session.getState() == ko.dbg.manager.STATE_RUNNING) {
            // cannot do stack for running threads
            return;
        }
        var old = this.debuggerPanel.callstackListbox.getItemAtIndex(this.currentStackFrameIndex);
        old.removeAttribute('currentview');

        this.currentStackFrameIndex = index;
        this.doStackFrame();

        this.debuggerPanel.callstackListbox.selectedItem.setAttribute('currentview','true');
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.doStackFrame = function DSTM_doStackFrame()
{
    log.debug("DebugSessionTabManager["+this.id+"].doStackFrame("+this.currentStackFrameIndex+")");
    try {
        // rebuild the variable lists
        this.updateVariables();
        this.session.showStackPosition(this.stackFrames[this.currentStackFrameIndex],
                                       this.currentStackFrameIndex);
    } catch(ex) {
        log.exception(ex);
    }
}

/* these functions provide state for variable viewers between stack
   changes and sessions in the same tab.  Currently they only store
   whether the variable is opened or not in the viewers. */
DebugSessionTabManager.prototype.getVarState = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].getVarState("+expression+")");
    try {
        if (typeof(this._varState[expression]) != 'undefined')
            return this._varState[expression];
    } catch(ex) {
        log.exception(ex);
    }
    return 0;
}

DebugSessionTabManager.prototype.contractVar = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].contractVar("+expression+")");
    try {
        this._varState[expression] = 0
    } catch(ex) {
        log.exception(ex);
    }
}

DebugSessionTabManager.prototype.expandVar = function(expression)
{
    log.debug("DebugSessionTabManager["+this.id+"].expandVar("+expression+")");
    try {
        this._varState[expression] = 1
    } catch(ex) {
        log.exception(ex);
    }
}

// Method template:
//DebugSessionTabManager.prototype.foobar = function()
//{
//    log.debug("DebugSessionTabManager["+this.id+"].foobar()");
//    try {
//    } catch(ex) {
//        log.exception(ex);
//    }
//}


//---- module (i.e. not tab-specific) interface routines

/**
 * The per-komodo-window instance of outputTabManager
 */
this.tabManager = null;

this.tabInit =
function DebugSessionTab_OnLoad()
{
    log.debug("DebugSessionTab_OnLoad()");
    try {
        ko.dbg.tabManager = new outputTabManager();
    } catch(ex) {
        log.exception(ex);
    }
}

function outputTabManager() {
}
outputTabManager.prototype = {
    tabManagers: {},

    // Return the current debug state in string format.
    getStatusAsString: function(state)
    {
        switch(state) {
        case ko.dbg.manager.STATE_STARTING:
            return "starting";
        case ko.dbg.manager.STATE_RUNNING:
            return "running";
        case ko.dbg.manager.STATE_BREAK:
            return "break";
        case ko.dbg.manager.STATE_INTERACTIVE:
            return "interactive";
        case ko.dbg.manager.STATE_STOPPING:
            return "stopping";
        case ko.dbg.manager.STATE_STOPPED:
            return "stopped";
        }
        return '';
    },

    // Uncollapsed an existing debug session tab and tabpanel for use.
    //
    //  "id" is the unique id (it is a number) for the tab.
    //
    uncollapse: function(id)
    {
        log.debug("outputTabManager.uncollapse("+id+")");
        try {
            var tabId = "debugsession"+id+"-tab";
            var tabPanelId = "debugsession"+id+"-tabpanel";

            var tab = document.getElementById(tabId);
            var tabPanel = document.getElementById(tabPanelId);

            if (!tab || !tabPanel) {
                throw new Error("Could not uncollapsed debug session tab "+id+": elements '"+
                      tabId+"' and/or '"+tabPanelId+"' do not exist.");
            }

            if (tab.hasAttribute("collapsed"))
                tab.removeAttribute("collapsed");
            if (tabPanel.hasAttribute("collapsed"))
                tabPanel.removeAttribute("collapsed");
        } catch(ex) {
            log.exception(ex);
        }
    },

    // Create a new debug session tab.
    //
    //  "id" is the unique id (it is a number) for the tab.
    //
    createTab: function outputTabManager_createTab(id)
    {
        log.debug("outputTabManager.createTab()");
        try {
            var panel = document.createElementNS(NS_XUL, "browser");
            var panelDocument = null;

            /**
             * Handle when the DOM document gets created so we can inject the
             * necessary globals a pane would expect before any script runs
             */
            panel.addEventListener("DOMWindowCreated", (function(event) {
                if (event.originalTarget.defaultView.frameElement != panel) {
                    return; // this isn't the right window
                }
                panel.contentWindow.parent = window;
                ["ko", "xtk", "gEditorTooltipHandler"].forEach(function(k) {
                    panel.contentWindow[k] = window[k];
                });

                // need to use the load event, DOMWindowCreated happens too
                // early - that fires on some sort of synthetic about:blank
                // but not the actual document we want (because it just got
                // carried over somehow)
                panel.contentWindow.addEventListener("load", (function(event) {
                    if (event.originalTarget.defaultView.frameElement != panel) {
                        return; // not the right window
                    }
                    // make sure we only hook up the listener after manually
                    // setting up the selected tab, though, because at this
                    // point the tab isn't initialized yet.
                    panelDocument = panel.contentDocument;
                    panel.tabbox.selectedTab = panel.tab;
                    panelDocument.addEventListener("ko-widget-move-completed",
                                                   widgetMoveListener, false);
                    // we need to listen to both "select" on the tabstrip and
                    // "command" on the tab, to cover both select-by-dragging
                    // and select-by-clicking-on-a-different-tabstrip
                    panel.tab.parentNode.addEventListener("select",
                                                          tabSelectListener,
                                                          false);
                    panel.tab.addEventListener("command",
                                               tabSelectListener,
                                               false);
                }).bind(this), false);
    
            }).bind(this), false);

            /**
             * Listener for when a tab is selected
             */
            var tabSelectListener = (function ko_dbg_createTab_tabSelectListener(event) {
                var view = panelDocument.defaultView;
                var tab = view && view.frameElement.tab;
                if (!tab) {
                    // this happens when the panel's gone (but its containing
                    // tabstrip isn't); remove the event listener and don't do
                    // anything else.
                    event.target.removeEventListener(event.type,
                                                     tabSelectListener,
                                                     false);
                    return;
                }
                var eventTab = event.target;
                if ("selectedItem" in event.target) {
                    // this is a tabstrip
                    eventTab = event.target.selectedItem
                }
                if (eventTab != tab) {
                    // something else is selected
                    return;
                }
                this.tabSwitch(tab);
            }).bind(this);

            /**
             * Listener for when the widget is moved from one ko-pane to another
             */
            var widgetMoveListener = (function ko_dbg_createTab_widgetMoveListener(event) {
                event.getData("from-pane").tabs
                     .removeEventListener("select",
                                          tabSelectListener,
                                          false);
                event.getData("to-pane").tabs
                     .addEventListener("select",
                                       tabSelectListener,
                                       false);
                panelDocument.defaultView.frameElement.tab
                             .addEventListener("command",
                                               tabSelectListener,
                                               false);
            }).bind(this);

            panel.setAttribute("type", "ko-widget");
            panel.setAttribute('id', 'debugsession'+id+'-tabpanel');
            document.getElementById('workspace_bottom_area').addWidget(panel);
            panel.setAttribute("src",
                               "chrome://komodo/content/debugger/debuggerPanel.xul");
            panel.tab.setAttribute('id','debugsession'+id+'-tab');
            panel.setAttribute('tab_id','debugsession'+id+'-tab');
            panel.setAttribute('tabid',id);
            panel.setAttribute('label','Output');
            panel.setAttribute('crop', 'center');
            panel.setAttribute('dbg_status',
                               this.getStatusAsString(ko.dbg.manager.STATE_STOPPED));
            panel.setAttribute('tab_class','dbg_status_icon');
            panel.tab.setAttribute('class','dbg_status_icon');
        } catch(ex) {
            log.exception(ex);
        }
    },

    /* Give me the next tabMgr created for a particular language (in interactive
      mode) that is still active (start looking after the currently selected
      tab if the tab group has focus).*/
    getInteractiveShell: function(language)
    {
        try {
            for (var id = 0; id < MAX_DEBUG_SESSION_TABS; ++id) {
                // Skip over free tabs, tabs w/ no sessions, or non-interactive
                // tabs.
                if (typeof(this.tabManagers[id]) == "undefined" ||
                    this.tabManagers[id].session == null ||
                    ! this.tabManagers[id].isInteractive ||
                    this.tabManagers[id].session.language != language) {
                    continue;
                }
                return this.tabManagers[id];
            }
        } catch (e) {
            log.exception(e);
        }
        return null;
    },

    // Return a usable (not busy, cleared) Debug Session Tab manager.
    /**
     * Get a usable (not busy, cleared) Debug Session Tab manager
     * @param onSuccessCallback [optional] callback on success; takes one
     *      parameter, the tab manager.
     * @param onErrorCallback [optional]
     */
    getTabMgr: function outputTabManager_getTabMgr(onSuccessCallback, onErrorCallback)
    {
        log.debug("outputTabManager.getTabMgr()");
        if (!onSuccessCallback) onSuccessCallback = function() {};
        if (!onErrorCallback) onErrorCallback = function() {};

        try {
            // Find a free tab id.
            for (var id = 0; id < MAX_DEBUG_SESSION_TABS; ++id) {
                if (typeof(this.tabManagers[id]) == "undefined") {
                    // This is a free tab id, we will dynamically create the
                    // tab and tabpanel and use that.
                    break;
                } else if (!("session" in this.tabManagers[id])) {
                    // This tab is still half-way through initializing. Don't
                    // reuse this tab, that's a bad idea
                    // -- no op --
                } else if (this.tabManagers[id].session === null) {
                    // Manager slot is null, there is a collapsed debug session
                    // tab and tabpanel that can be used.
                    this.uncollapse(id);
                    onSuccessCallback(this.tabManagers[id]);
                    return;
                } else {
                    // Manager slot is non-null (presumably a
                    // DebugSessionTabManager): leave it alone.
                }
            }
            if (id >= MAX_DEBUG_SESSION_TABS) {
                ko.dialogs.alert("Too many debug session tabs are in use. You must "+
                             "close one or more debug session tabs and restart "+
                             "this debug session.");
                onErrorCallback(null);
                return;
            }

            // If we get here, we are in the free tab id case above
            this.createTab(id);
            // Create a manager for this tab and return it.
            var manager = new DebugSessionTabManager(id);
            this.tabManagers[id] = manager;
            var onPanelLoad = (function onPanelLoad_() {
                manager.tabPanel.contentWindow
                       .removeEventListener("load", onPanelLoad, false);
                manager.initialize(id);
                onSuccessCallback(manager);
            }).bind(this);
            manager.tabPanel.contentWindow
                   .addEventListener("load", onPanelLoad, false);
        } catch(ex) {
            log.exception(ex);
            onErrorCallback(ex);
        }
    },

    /* we just want to create a new tab, but not attach a
       manager to it */
    /**
     * Create a new tab, without attaching a manager to it
     * @param tabMgr [optional] the tab manager
     * @param onSuccessCallback [optional] the callback on success
     * @param onErrorCallback [optional]
     */
    showNewTab: function outputTabManager_showNewTab(tabMgr, onSuccessCallback, onErrorCallback)
    {
        log.debug("outputTabManager.showNewTab()");
        if (!onSuccessCallback) onSuccessCallback = function() {};
        if (!onErrorCallback) onErrorCallback = function() {};

        try {
            if (!tabMgr) {
                this.getTabMgr(onTabMgrReady.bind(this),
                               onErrorCallback);
            } else {
                onTabMgrReady.call(this, tabMgr);
            }
        } catch(ex) {
            log.exception(ex);
            onErrorCallback(ex);
        }

        function onTabMgrReady(tabMgr) {
            try {
                tabMgr.configure();
                tabMgr.show();
                ko.dbg.manager.setCurrentSession(null);
                this.releaseTabMgr(tabMgr);
            } catch (ex) {
                log.exception(ex);
                onErrorCallback(ex);
                return;
            }
            onSuccessCallback();
        }
    },

    releaseTabMgr: function(tabMgr)
    {
        // nothing to do now
    },

    /**
     * Return the active DebugSessionTabManager
     * 
     * @returns {DebugSessionTabManager} The active tab
     */
    getCurrentTab: function()
    {
        //log.debug("outputTabManager.getCurrentTab()");
        try {
            var session = ko.dbg.manager.currentSession;
            if (!session) {
                return null;
            }
            return this.tabManagers[session.panel.id] || null;
        } catch(ex) {
            log.exception(ex);
        }
        return null;
    },

    /**
     * Close a given tab
     * @param tab the tab to close, or its linked panel
     */
    closeTab: function outputTabManager_closeTab(tab) {
        if (!tab) {
            log.exception("closeTab called without a tab!");
            throw "closeTab called without a tab";
        }
        if (tab.frameElement && tab.frameElement.tab) {
            // this is a window inside the tab, I think?
            tab = tab.frameElement;
        }
        if (tab.tab) {
            // this is a linked panel
            tab = tab.tab;
        }
        try {
            var id = tab.linkedpanel.getAttribute('tabid');
            // if this is not a managed tab (eg. command output)
            // then it will not have a tabid attribute
            if (!id) {
                log.debug(" attempt to close tab without id");
                return;
            }
            var tabMgr = this.tabManagers[id];
            if (!tabMgr) return;

            // XXX we could call a canclose function here, and validate
            // that the user wants to shut down the debug session,
            // for now, the easy way out.
            if (!tabMgr.canClose()) {
                return;
            }

            delete this.tabManagers[id];
            tabMgr.forceClose();
            var widgetId = "debugsession"+id+"-tabpanel"
            var widget = document.getElementById(widgetId);
            var tabbox = widget.tabbox;
            tabbox.removeWidget(widget);
            for each (tabMgr in this.tabManagers) {
                if (!("session" in tabMgr)) {
                    // not completely initialized yet
                    continue;
                }
                if (!tabMgr.session) {
                    // no attached session?
                    continue;
                }
                tabMgr.tabPanel.tabbox.selectedTab = tabMgr.tab;
                this.tabSwitch(tabMgr.tab);
                break;
            }
        } catch(ex) {
            log.exception(ex);
        }
        
    },

    closeCurrent: function()
    {
        log.debug("outputTabManager.closeCurrent()");
        try {
            var tab = this.getCurrentTab();
            if (tab) {
                this.closeTab(tab);
            }
        } catch(ex) {
            log.exception(ex);
        }
    },

    tabSwitch: function(tab)
    {
        log.debug("outputTabManager.tabSwitch("+tab.id+")");
        try {
            var id = tab.linkedpanel.getAttribute('tabid');
            // if this is not a managed tab (eg. command output)
            // then it will not have a tabid attribute
            if (!id || typeof(this.tabManagers[id]) == 'undefined') return;
            for each (var manager in this.tabManagers) {
                if (manager != this.tabManagers[id]) {
                    manager.tab.removeAttribute("current-session");
                }
            }
            tab.setAttribute("current-session", true);
            ko.dbg.manager.setCurrentSession(this.tabManagers[id].session);
        } catch(ex) {
            log.exception(ex);
        }
    }

}

}).apply(ko.dbg);
